from kmerpapa.pattern_utils import *
from kmerpapa.CV_tools import make_all_folds_contextD_kmers
from kmerpapa.score_utils import get_betas
import sys
import numpy
from scipy.special import xlogy, xlog1py

'''
Handles cross validation.
Should model-selection be determined based on cross validation?
Do I need backtrack for the Cross-validation?
Maybe i just need CV to estimate best meta-par value
And then return best papa for all data using that meta-par.
'''

def score_folds(trainM, trainU, alphas, betas, penalty):
    p = (trainM + alphas)/(trainM + trainU + alphas + betas)
    res = -2*(xlogy(trainM, p) + xlog1py(trainU,-p)) + penalty
    if res<0:
        return 1e100
    return res

def test_folds(trainM, trainU, testM, testU, alphas, betas):
    '''
    likelihood of test_data on training set
    '''
    #p = (trainM + alphas -1)/(trainM + trainU + alphas + betas -2)
    p = (trainM + alphas)/(trainM + trainU + alphas + betas)
    return -2 *(xlogy(testM, p) + xlog1py(testU, -p))


def get_score(pattern, contextD, alpha, beta, penalty):
    M,U = get_M_U(pattern, contextD)
    return score_folds(M, U, alpha, beta, penalty)

def get_test_train(pattern, trainD, testD,  alphas, betas):
    M_train,U_train = get_M_U(pattern, trainD)
    M_test, U_test = get_M_U(pattern, testD)
    train_score = test_folds(M_train, U_train, M_train, U_train, alphas, betas)
    test_score = test_folds(M_train, U_train, M_test, U_test, alphas, betas)
    return train_score, test_score

def greedy_res(pattern, contextD, alphas, betas, penalty):
    if pattern in contextD:
        return get_score(pattern, contextD, alphas, betas, penalty), [pattern]
    else:
        best_score = get_score(pattern, contextD, alphas, betas, penalty)
        best_papa = [pattern]
        for i in range(len(pattern)):#[::-1]:
            if pattern[i] not in complements:
                continue
            for c1, c2 in complements[pattern[i]]:
                pat1 = pattern[:i] + c1 + pattern[i+1:]
                pat2 = pattern[:i] + c2 + pattern[i+1:]

                score1 = get_score(pat1, contextD, alphas, betas, penalty)
                score2 = get_score(pat2, contextD, alphas, betas, penalty)

                s = score1 + score2
                if s < best_score:
                    best_score = s
                    best_papa = [pat1, pat2]

        if len(best_papa) > 1:
            pat1, pat2 = best_papa
            #print('split', pat1, pat2)
            s1, papa1 = greedy_res(pat1, contextD, alphas, betas, penalty)
            s2, papa2 = greedy_res(pat2, contextD, alphas, betas, penalty)
            return s1+s2, papa1+papa2
        else:
            return best_score, best_papa

def greedy_partition_CV(gen_pat, contextD, alphas, args, nmut, nunmut, penalties, index_mut=0):
    resD = {}
    nf = args.nfolds
    nit = args.iterations
    npat = generality(gen_pat)
    U_mem = numpy.zeros((npat,nf), dtype=numpy.uint64)
    M_mem = numpy.zeros((npat,nf), dtype=numpy.uint64)
    M_sum_test = numpy.zeros(nf)
    U_sum_test = numpy.zeros(nf)
    test_loss = {(a_i, p_i):[] for a_i in range(len(alphas)) for p_i in range(len(penalties))}
    train_loss = {(a_i, p_i):[] for a_i in range(len(alphas)) for p_i in range(len(penalties))}
    #{(a_i, p_i):[] for a in alphas for p in penalties}
    prng = np.random.RandomState(args.seed)
    for iteration in range(nit):
        if args.verbosity > 0:
            print(f'greedy CV iteration {iteration}', file=sys.stderr)
        make_all_folds_contextD_kmers(contextD, U_mem, M_mem, gen_pat, prng)
        M_sum_test = M_mem.sum(axis=0) # n_mut for each fold
        U_sum_test = U_mem.sum(axis=0) # n_unmut for each fold 
        M_sum_train = sum(M_sum_test) - M_sum_test
        U_sum_train = sum(U_sum_test) - U_sum_test
        p_all = M_sum_test/(M_sum_test + U_sum_test)
        LL_intercept = -2*(xlogy(M_sum_test, p_all) + xlog1py(U_sum_test, -p_all))

        trainDs = []
        testDs = []
        for fold in range(nf):
            trainDs.append({})
            testDs.append({})

        for i, context in enumerate(matches(gen_pat)):
            M_test = M_mem[i]
            U_test = U_mem[i]
            M_train = sum(M_test) - M_test
            U_train = sum(U_test) - U_test
            for fold in range(nf):
                trainDs[fold][context] = (M_train[fold], U_train[fold])
                testDs[fold][context] = (M_test[fold], U_test[fold])

        for a_i in range(len(alphas)):
            alpha = alphas[a_i]
            for p_i in range(len(penalties)):
                penalty = penalties[p_i]
                betas = get_betas(alpha, M_sum_train, U_sum_train)
                test_score = [0] * nf
                train_score = [0] * nf
                for fold in range(nf):
                    best_score, papa = greedy_res(gen_pat, trainDs[fold], alpha, betas[fold], penalty)
                    for pattern in papa:
                        train_s, test_s = get_test_train(pattern, trainDs[fold], testDs[fold],  alpha, betas[fold])
                        train_score[fold] += train_s
                        test_score[fold] += test_s
                if args.verbosity > 0:
                    print(f'CV on k={len(gen_pat)} alpha={alpha} penalty={penalty} i={iteration} test_LL={sum(test_score)}', file=sys.stderr)
                if args.verbosity > 1:
                    print(f'test LL for each fold: {test_score}', file=sys.stderr)
                train_loss[(a_i, p_i)].extend(train_score)
                test_loss[(a_i, p_i)].extend(test_score)
    best_test_loss = 1e100
    best_values = (None, None)
    for a_i in range(len(alphas)):
        alpha = alphas[a_i]
        for p_i in range(len(penalties)):
            penalty = penalties[p_i]
            test = sum(test_loss[(a_i, p_i)])/nit
            if not args.CVfile is None:    
                print(len(gen_pat), alpha, penalty, test, file = args.CVfile)
            if test < best_test_loss:
                best_values = (alpha, penalty)
                best_test_loss = test
    return best_values[0], best_values[1], best_test_loss

def greedy_partition(gen_pat, contextD, alpha, beta, penalty, args):
    best_score, papa = greedy_res(gen_pat, contextD, alpha, beta, penalty)
    M, U = get_M_U(gen_pat, contextD)
    return best_score, M, U, papa




"""Module that contains the command line application."""

import argparse
import sys
from kmerpapa.pattern_utils import LCA_pattern_of_kmers, get_M_U
from kmerpapa.score_utils import get_loss
from kmerpapa.papa import *
from kmerpapa.io_utils import read_input, downsize_contextD
from math import log    
import kmerpapa.algorithms.all_kmers_CV
from kmerpapa.algorithms import bottum_up_array_penalty_plus_pseudo_CV
from kmerpapa.algorithms import greedy_penalty_plus_pseudo
from kmerpapa.algorithms import bottum_up_array_w_numba

def get_parser():
    """
    Return the CLI argument parser.

    Returns:
        An argparse parser.
    """
    parser = argparse.ArgumentParser(
        prog="kmerpapa",
        description='Finds optimal k-mer pattern partition in fx. mutation data')
    parser.add_argument(
        '-p', '--positive', type=argparse.FileType('r'),
        help='File with k-mer counts in positive set')
    parser.add_argument(
        '-n', '--negative', type=argparse.FileType('r'),
        help='File with k-mer counts in negative set. '
        'If the negative set is created with a larger k than the positive set then the k-mers '
        'will be collapsed so that they have the same length.')
    parser.add_argument(
        '-b', '--background', type=argparse.FileType('r'),
        help='File with k-mer counts in backgound set (includes both positive and negative regions). '
        'If the background set is created with a larger k than the positive set then the k-mers '
        'will be collapsed so that they have the same length.')
    parser.add_argument(
        '-j', '--joint_context_counts', type=argparse.FileType('r'),
        help='File with k-mer counts in positive set and background set. '
        'This option can be used instead of having positive and '
        'negative counts in seperate files.')
    ## It is probably best not to use the --scale_factor option but instead just scale the
    ## rates afterwards. So I am commenting it out for now. If I want to use it again I need
    ## to use args.scale_factor in io_utils.read_input again.
    #parser.add_argument(
    #    '-m', '--scale_factor', default='1', type=float,
    #    help='All background (or negative) counts will be multiplied by this number. '
    #    'If the positive set is based on observations from n genomes and the bacgkround '
    #    'is based on the number of occurances in the reference genome then the scale factor '
    #    'should be 2*n.')
    parser.add_argument(
        '-o', '--output', type=argparse.FileType('w'), default='-',
        metavar='PATH',
        help="Output file (default: standard output)")
    parser.add_argument(
        '-f', '--CVfile', type=argparse.FileType('w'),
        help='File with training and test likelihood values from cross validation.')
    parser.add_argument(
        '--verbosity', type = int, default=1,
        help="Amount of info printed to stderr during execution. 0:silent, 1:default, 2:verbose")
    parser.add_argument(
        '--CV_only', action='store_true',
        help="Only run crossvalidation. Do not run on whole data set using best values afterwards.")
    parser.add_argument(
        '--greedy', action='store_true',
        help="Use a fast greedy heuristic to find a (hopefully) good but not necessarily optimal pattern partition.")
    parser.add_argument(
        '--greedyCV', action='store_true',
        help="Use a greedy heuristic during CV but use optimal algorithm afterwards")
    parser.add_argument(
        '-l', '--long_output', action='store_true',
        help="Print all k-mers in output format.")
    parser.add_argument(
        '-s', '--super_pattern', type=str,
        help='If a super-pattern is provided the program will only consider k-mers that match that pattern. '
             'If for instance the "--positive" file contain all 5-mers at A->T mutated sites but the "--background" '
             'file contains 5-mers from all sites in the genome. Then "--super_pattern NNANN" should be specified to '
             'ignore 5-mers where A->T mutations cannot happen.')
    parser.add_argument(
        '--score', type=str, default='penalty_and_pseudo', choices=['penalty_and_pseudo', 'all_kmers', 'BIC', 'AIC', 'HQ', 'LL'],
        help='Type of score function. Default is "penalty_and_pseudo". '
             '"all_kmers" will calculate a rate for each k-mer.')
    parser.add_argument(
        '-N', '--nfolds', type=int, metavar='N',
        help='Perform cross validation with N folds. '
             'If more than one value of pseudo_count and penalty is given then default is 2. '
             'Otherwise default is not to run cross validation if --nfolds option is not set.')
    parser.add_argument(
        '-i', '--iterations', type=int, default=1, metavar='i',
        help='Repeat cross validation i times')
    parser.add_argument(
        '-a', '--pseudo_counts', type=float, metavar='a', nargs='+', default = [0.8],
        help='Different pseudo count (alpha) values to test using cross validation')
    parser.add_argument(
        '-c', '--penalty_values', type=float, metavar='c', nargs='+',
        help='Different penalty values to test using cross validation. '
             'If no value is set for the default scoring function then '
             'log(#k-mers) will be used.')
    parser.add_argument(
        '--test_smaller_k', action='store_true',
        help='By standard k is the width of the k-mers in the input data. '
        'If this option is supplied it will test all odd numbern up to the width using CV '
        'and use the best.')
    parser.add_argument(
        '--seed', type=int,
        help='seed for numpy.random')
    parser.add_argument(
        '-V', '--version', action='store_true',
        help="Print version number and return")
    return parser


def main(args = None):
    """
    Run the main program.

    This function is executed when you type `kmerpapa` or `python -m kmerpapa`.

    Arguments:
        args: Arguments passed from the command line.

    Returns:
        An exit code.
    """
    parser = get_parser()
    args = parser.parse_args(args=args)

    if args.version:
        from kmerpapa import __version__
        print("version:", __version__)
        print()
        return 0

    if not args.super_pattern is None:
        super_pattern = Pattern(args.super_pattern)
    else:
        super_pattern = None

    try:
        contextD, n_unmut, n_mut = read_input(args, super_pattern)
    except Exception as e:
        parser.print_help()
        print('='*80, file=sys.stderr)
        print("input error:", file=sys.stderr)
        print(e, file=sys.stderr)
        print('='*80, file=sys.stderr)

        return 0

    if args.verbosity > 0:
        print(f'Input data read. {n_mut} positive k-mers and {n_unmut} negative k-mers', file=sys.stderr)

    if not args.penalty_values is None:
        assert args.score == 'penalty_and_pseudo', f'you cannot specify penalty values when using the {args.score} score function'
    else:
        if args.score == "BIC":
            args.penalty_values = [log(n_mut)]
        elif args.score == "AIC":
            args.penalty_values = [2.0]
        elif args.score == "HQ":
            args.penalty_values = [log(log(n_mut))]
        elif args.score == "LL":
            args.penalty_values = [0.0]
        elif args.score == 'all_kmers':
            pass
        elif args.score == 'penalty_and_pseudo':
            args.penalty_values = [log(len(contextD))]
            if args.verbosity > 0:
                print(f'penalty values not set. Using {args.penalty_values[0]}', file=sys.stderr)
        else:
            assert False, f"illegal score option {args.score}"

 
    gen_pat = LCA_pattern_of_kmers(list(contextD.keys()))

    if not args.super_pattern is None:
        assert gen_pat == args.super_pattern

    for context in matches(gen_pat):
        if context not in contextD:
            contextD[context] = (0,0)

    if args.verbosity > 0:
        print(f'General pattern: {gen_pat}', file=sys.stderr)

    if not args.CVfile is None:
        print('k alpha P LL_test', file=args.CVfile)

    best_alpha = None
    best_penalty = None
    best_k = None

    if args.test_smaller_k:
        ks = range(len(gen_pat),1,-2)
    else:
        ks = [len(gen_pat)]

    this_contextD = contextD
    this_gen_pat = gen_pat
    best_score = 1e100

    if args.nfolds is None and (len(ks) > 1 or len(args.pseudo_counts) > 1 or len(args.penalty_values) > 1 or args.CV_only):
        args.nfolds = 2            

    if not args.nfolds is None and args.nfolds > 1:
        for k in ks:
            if args.verbosity > 0:
                print(f'Running {args.nfolds}-fold cross validation on {k}-mers', file=sys.stderr)
            if k != len(this_gen_pat):
                this_contextD, this_gen_pat = downsize_contextD(this_contextD, this_gen_pat, k)
            if args.greedy or args.greedyCV:
                assert args.score != 'all_kmers', 'greedy option cannot be used wil all-kmers'
                this_alpha, this_penalty, test_score = greedy_penalty_plus_pseudo.greedy_partition_CV(this_gen_pat, this_contextD, args.pseudo_counts, args, n_mut, n_unmut, args.penalty_values)
            elif args.score == 'all_kmers':
                this_alpha, test_score =  kmerpapa.algorithms.all_kmers_CV.all_kmers(this_gen_pat, this_contextD, args.pseudo_counts, args, n_mut, n_unmut)
                this_penalty = None
            else:
                this_alpha, this_penalty, test_score = bottum_up_array_penalty_plus_pseudo_CV.pattern_partition_bottom_up(this_gen_pat, this_contextD, args.pseudo_counts, args, n_mut, n_unmut, args.penalty_values)
            if test_score < best_score:
                best_score = test_score
                best_k = k
                best_alpha = this_alpha
                best_penalty = this_penalty
        if args.verbosity > 0:
            print(f'CV DONE. best_k={best_k}, best_alpha={best_alpha}, best_penalty={best_penalty}, best_test_LL={best_score}', file=sys.stderr)

    if not args.CVfile is None:
        args.CVfile.close()

    if args.CV_only:
        return 0
    
    if best_alpha is None:
        assert len(args.pseudo_counts)==1
        best_alpha = args.pseudo_counts[0]

    if args.score != 'all_kmers' and best_penalty is None:
        assert len(args.penalty_values)==1
        best_penalty = args.penalty_values[0]

    if best_k is None:
        best_k = len(gen_pat)

    if best_k != len(gen_pat):
        contextD, gen_pat = downsize_contextD(contextD, gen_pat, best_k)

    my=n_mut/(n_mut+n_unmut)
    best_beta = (best_alpha*(1.0-my))/my

    if args.verbosity >0:
        print(f'Training on whole data set with k={best_k} alpha={best_alpha} penalty={best_penalty}' ,file=sys.stderr)

    if args.score == 'all_kmers':
        #TODO: maybe I should calculate best score also for all_kmers
        best_score = 0
        M = n_mut
        U = n_unmut
        names = list(matches(gen_pat))
    elif args.greedy:
        best_score, M, U, names = \
            greedy_penalty_plus_pseudo.greedy_partition(gen_pat, contextD, best_alpha, best_beta, best_penalty, args)
    else:
        best_score, M, U, names = \
            bottum_up_array_w_numba.pattern_partition_bottom_up(gen_pat, contextD, best_alpha, best_beta, best_penalty, args, n_mut, n_unmut)
        
    counts = []
    for pat in names:
        counts.append(get_M_U(pat, contextD))

    # Just to check that it is a partition
    #sp = PatternPartition(list(names), superPattern=gen_pat)
    # Check removed because it takes too long on large k datasets with all_kmers.
    #print(M, n_mut)
    assert M == n_mut
    assert U == n_unmut
    assert n_mut == sum(x[0] for x in counts)
    assert n_unmut == sum(x[1] for x in counts)

    total_rate = float(n_mut)/(n_mut+n_unmut)

    if args.verbosity>0:
        print(f'Optimal k-mer pattern partition contains {len(names)} patterns.', file=sys.stderr)
        print(f'loss={best_score}', file=sys.stderr)
        print(f'LL={get_loss(counts, best_alpha, best_beta)}', file=sys.stderr)

    if args.long_output:
        print('context', 'c_neg', 'c_pos', 'c_rate',
                'pattern', 'p_neg', 'p_pos', 'p_rate', file=args.output)
    else:
        print('pattern', 'p_neg', 'p_pos', 'p_rate', file=args.output)

    for i in range(len(names)):
        pat = names[i]
        M, U = counts[i]
        p = (M + best_alpha)/(M + U + best_alpha + best_beta)
        if args.long_output:
            for context in matches(pat):
                nm, ns = contextD[context]
                print(context, ns, nm, float(nm)/(nm+ns), pat, U, M, p, file=args.output)
        else:
            print(pat, U, M, p, file=args.output)

    return 0



from kmerpapa.pattern_utils import *
from kmerpapa.score_utils import get_betas
import kmerpapa.CV_tools
import sys
from numba import njit
import numpy as np
import scipy.special

'''
Handles cross validation. Used to find best meta-parameter values.
Back-track is not implemented for CV. After best meta-parameters are found kmerpapa.bottom_up_array_w_numba 
should be used to train model on all data and to backtrack to find best solution
'''

def score_test_folds(trainM, trainU, testM, testU, alphas, betas, penalty):
    p = (trainM + alphas)/(trainM + trainU + alphas + betas)
    #TODO: log p udregnes to gange er det hurtigere at udregne det selv end at bruge xlogy?
    train_score = -2.0*(scipy.special.xlogy(trainM, p) + scipy.special.xlog1py(trainU,-p)) + penalty
    test_ll =  -2.0*(scipy.special.xlogy(testM, p) + scipy.special.xlog1py(testU, -p))
    return train_score, test_ll

@njit
def get_train(test):
    return test.sum() - test

@njit
def handle_pattern(score_mem, test_score_mem, pattern, M_mem, U_mem, alpha, betas, penalty):
    """Fill in dynamic programming tables for a pattern
    """
    pat_num = 0
    for i in range(gen_pat_len):
        pat_num += perm_code_no_np[gpo[i]][pattern[i]] * cgppl[i]
    res_train = score_mem[pat_num]
    res_test = test_score_mem[pat_num]

    first = True
    for i in range(gen_pat_len):
        if has_complement[pattern[i]]:
            pcn_row = perm_code_no_np[gpo[i]]
            pat_num_diff  = pat_num - pcn_row[pattern[i]] * cgppl[i]
            for j in range(n_complements[pattern[i]]):
                c1 = complements_tab_1[pattern[i]][j]
                c2 = complements_tab_2[pattern[i]][j]
                pat1_num = pat_num_diff + pcn_row[c1] * cgppl[i]
                pat2_num = pat_num_diff + pcn_row[c2] * cgppl[i]
                new_train = score_mem[pat1_num]+score_mem[pat2_num] 
                new_test = test_score_mem[pat1_num] + test_score_mem[pat2_num]
                for k in range(nf):
                    if new_train[k] < res_train[k]:
                        res_train[k] = new_train[k]
                        res_test[k] = new_test[k]
                if first:
                    M_mem[pat_num] = M_mem[pat1_num] + M_mem[pat2_num]
                    U_mem[pat_num] = U_mem[pat1_num] + U_mem[pat2_num]
                    first = False
    M_test = M_mem[pat_num]
    U_test = U_mem[pat_num]
    M_train = M_test.sum() - M_test 
    U_train = U_test.sum() - U_test 
    ps = (M_train + alpha)/(M_train + U_train + alpha + betas)
    logps = np.log(ps)
    log1mps = np.log(1.0-ps)
    for i in range(nf):
        logp = logps[i]
        log1mp = log1mps[i]
        s = penalty
        if M_train[i] > 0:
            s += -2.0 * M_train[i]*logp
        if U_train[i] > 0:
            s += -2.0 * U_train[i]*log1mp
        if s < res_train[i]:
            res_train[i] = s
            s = 0.0
            if M_test[i] > 0:
                s += -2.0 * M_test[i]*logp
            if U_test[i] > 0:
                s += -2.0 * U_test[i]*log1mp
            res_test[i] = s


def pattern_partition_bottom_up(gen_pat, contextD, alphas, args, nmut, nunmut, penalties, index_mut=0):
    global cgppl
    global gppl
    global gpo
    global gen_pat_len
    global has_complement
    global nf
    global gen_pat_num
    ftype = np.float32
    npat = pattern_max(gen_pat)
    nf = args.nfolds
    nit = args.iterations
    test_score_mem = np.empty((npat,nf), dtype=ftype)
    if nmut+nunmut > np.iinfo(np.uint32).max:
        itype = np.uint64
    else:
        itype = np.uint32

    PE = PatternEnumeration(gen_pat)

    U_mem = np.empty((npat,nf), dtype=itype)
    M_mem = np.empty((npat,nf), dtype=itype)
    #U_mem[2,1] indeholder antal test sampels i pat_no 2 foerste fold
    #test kan fåes ved at trække fra sum: sum(U_mem[pat])-U_mem[pat]

    cgppl = np.array(get_cum_genpat_pos_level(gen_pat), dtype=np.uint32)
    gppl = np.array(get_genpat_pos_level(gen_pat), dtype=np.uint32)
    gpo = np.array([ord(x) for x in gen_pat], dtype=np.uint32)

    gen_pat_level = pattern_level(gen_pat)
    gpot = tuple(ord(x) for x in gen_pat)

    test_loss = {(a_i, p_i):[] for a_i in range(len(alphas)) for p_i in range(len(penalties))}
    train_loss = {(a_i, p_i):[] for a_i in range(len(alphas)) for p_i in range(len(penalties))}
    gen_pat_len = len(gen_pat)
    gen_pat_num = PE.pattern2num(gen_pat)

    has_complement = np.full(100,True)
    has_complement[ord('A')] = False
    has_complement[ord('C')] = False
    has_complement[ord('G')] = False
    has_complement[ord('T')] = False

    prng = np.random.RandomState(args.seed)

    # nit is number of itereations if we do repeated CV
    for iteration in range(nit):
        if args.verbosity > 0 and nit > 1:
            print('CV Iteration', iteration, file=sys.stderr)
        kmerpapa.CV_tools.make_all_folds_contextD_patterns(contextD, U_mem, M_mem, gen_pat, prng, itype)
        if args.verbosity > 0:
            print('CV sampling DONE', file=sys.stderr)
    
        M_sum_test = M_mem.sum(axis=0) # n_mut for each fold
        U_sum_test = U_mem.sum(axis=0) # n_unmut for each fold 
        M_sum_train = M_sum_test.sum() - M_sum_test
        U_sum_train = U_sum_test.sum() - U_sum_test

        for a_i in range(len(alphas)):
            alpha = alphas[a_i]
            betas = get_betas(alpha, M_sum_train, U_sum_train)
            for p_i in range(len(penalties)):
                score_mem = np.full((npat,nf), 1e100, dtype=ftype)
                penalty = penalties[p_i]
                for pattern in subpatterns_level(gen_pat, 0):
                    pat_num = PE.pattern2num(pattern)
                    M_test = M_mem[pat_num]
                    U_test = U_mem[pat_num]
                    M_train = get_train(M_test)
                    U_train = get_train(U_test)
                    score_mem[pat_num], test_score_mem[pat_num] = score_test_folds(M_train, U_train, M_test, U_test, alpha, betas, penalty)

                for level in range(1, gen_pat_level+1):
                    if args.verbosity > 1:    
                        print(f'level {level} of {gen_pat_level}', file=sys.stderr)
                    for pattern in subpatterns_level_ord_np(gpot, gen_pat_level, level):
                        handle_pattern(score_mem, test_score_mem, pattern, M_mem, U_mem, alpha, betas, penalty)
                if args.verbosity > 0:
                    print(f'CV on k={len(gen_pat)} alpha={alpha} penalty={penalty} i={iteration} test_LL={sum(test_score_mem[gen_pat_num])}', file=sys.stderr)
                if args.verbosity > 1:
                    print(f'test LL for each fold: {test_score_mem[gen_pat_num]}', file=sys.stderr)
                train_loss[(a_i, p_i)].extend(list(score_mem[gen_pat_num]))
                test_loss[(a_i, p_i)].extend(list(test_score_mem[gen_pat_num]))

    best_test_loss = 1e100
    best_values = (None, None)
    for a_i in range(len(alphas)):
        alpha = alphas[a_i]
        for p_i in range(len(penalties)):
            penalty = penalties[p_i]
            test = sum(test_loss[(a_i, p_i)])/nit
            if not args.CVfile is None:    
                print(len(gen_pat), alpha, penalty, test, file = args.CVfile)
            if test < best_test_loss:
                best_values = (alpha, penalty)
                best_test_loss = test
    return best_values[0], best_values[1], best_test_loss

